#!/usr/bin/env python

import os

CODE_ANALYSIS=True
DEEP_CODE_ANALYSIS=False
CONFIG_ANALYSIS_TIMEOUT=0
PLUGINS_PATH=os.path.dirname(__file__) + os.sep + "plugins"
DATABASE_PATH=os.path.dirname(__file__) + os.sep + "files.sqlite"
